/*
 * T-Plan Robot Enterprise, automated testing tool.
 * Copyright (C) 2009-2013 T-Plan Limited (http://www.t-plan.com),
 * Cloisters House, 8 Battersea Park Road, London, SW8 4BG, United Kingdom.
 * All rights reserved.
 */
package com.tplan.adb;

import com.android.ddmlib.FileListingService;
import com.android.ddmlib.FileListingService.FileEntry;
import com.android.ddmlib.IDevice;
import com.tplan.robot.ApplicationSupport;
import com.tplan.robot.util.Utils;
import java.awt.AlphaComposite;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * A simple file viewer window allowing to browse the file system on the Android
 * device over the ADB connection.
 */
class AndroidFileViewer extends JDialog implements ActionListener, Runnable, MouseListener, ListSelectionListener {

    IDevice device;
    JList list = new JList();
    JButton btnOk = new JButton("Choose", ApplicationSupport.getImageIcon("ok15.gif"));
    Component owningButton;
    private String selectedFile;
    private String fileToSelect;
    FileItem currentRoot;
    private static Icon dirIcon;
    private static Icon fileIcon;

    AndroidFileViewer(Component owner, IDevice device) {
        super(SwingUtilities.getWindowAncestor(owner), ModalityType.APPLICATION_MODAL);
        this.device = device;
        this.owningButton = owner;
        setTitle(device.getName() + " File System");

        // Initialize the components
        JPanel pnl = new JPanel(new BorderLayout());
        pnl.add(new JScrollPane(list), BorderLayout.CENTER);
        JPanel south = new JPanel(new FlowLayout(FlowLayout.CENTER));
        btnOk.setEnabled(false);
        btnOk.setActionCommand("ok");
        btnOk.addActionListener(this);
        south.add(btnOk);
        JButton btnCancel = new JButton("Cancel", ApplicationSupport.getImageIcon("cancel16.png"));
        btnCancel.setActionCommand("cancel");
        btnCancel.addActionListener(this);
        south.add(btnCancel);
        pnl.add(south, BorderLayout.SOUTH);
        setContentPane(pnl);
        setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
        Utils.registerDialogForEscape(this, btnCancel);
        getRootPane().setDefaultButton(btnOk);

        list.setVisibleRowCount(12);
        list.setEnabled(false);
        DefaultListModel model = new DefaultListModel();
        model.add(0, "Wait while retrieving files...");
        list.setModel(model);
        list.addListSelectionListener(this);
        list.addMouseListener(this);
        initIcons();
        list.setCellRenderer(new FileRenderer());
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    }

    private void initIcons() {
        if (dirIcon == null) {
            dirIcon = UIManager.getIcon("FileView.directoryIcon");
            if (dirIcon == null) {
                dirIcon = ApplicationSupport.getImageIcon("open16.gif");
            }
        }
        if (fileIcon == null) {
            fileIcon = UIManager.getIcon("FileView.fileIcon");
            // Create a transparent empty icon of the same size as the dir one
            // to ensure that file names will be indented appropriately
            if (fileIcon == null) {
                BufferedImage img = new BufferedImage(dirIcon.getIconWidth(), dirIcon.getIconHeight(), BufferedImage.TYPE_INT_ARGB);
                Graphics2D g2 = img.createGraphics();
                g2.setComposite(AlphaComposite.getInstance(AlphaComposite.CLEAR));
                g2.fillRect(0, 0, dirIcon.getIconWidth(), dirIcon.getIconHeight());
                fileIcon = new ImageIcon(img);
            }
        }
    }

    public void actionPerformed(ActionEvent e) {
        if ("ok".equals(e.getActionCommand())) {
            selectedFile = ((FileItem) list.getSelectedValue()).entry.getFullEscapedPath();
            dispose();
        } else if ("cancel".equals(e.getActionCommand())) {
            selectedFile = null;
            dispose();
        }
    }

    @Override
    public void setVisible(boolean visible) {
        if (visible) {
            pack();
            Dimension d = getSize();
            d.width = Math.max(d.width, 420);
            setSize(d);
            setLocationRelativeTo(getParent());
            new Thread(this).start();
        }
        super.setVisible(visible);
    }

    static FileEntry getEntryForPath(IDevice device, String file) {
        String path[] = file.split("[/]");
        FileListingService fs = device.getFileListingService();
        FileEntry fe = fs.getRoot();

        FileEntry fl[];
        // Skip the first entry because it's empty
        String dir;
        for (int i = 1; i < path.length; i++) {
            dir = path[i];
            fl = fs.getChildren(fe, true, null);
            fe = null;
            if (fl != null && fl.length > 0) {
                for (FileEntry e : fl) {
                    if (e.getName().equals(dir)) {
                        fe = e;
                        break;
                    }
                }
            }
            if (fe == null) {
                return null;
            }
        }
        return fe;
    }

    public void run() {
        FileEntry temp = null;
        if (fileToSelect != null) {
            temp = getEntryForPath(device, fileToSelect);
            // Consume the file
            fileToSelect = null;
        }
        if (currentRoot == null) {
            FileEntry root = null;
            if (temp != null) {
                root = temp.getParent();
            }
            if (root == null) {
                root = device.getFileListingService().getRoot();
            }
            currentRoot = new FileItem(root, true);
        }
        final DefaultListModel model = populateList(currentRoot);
        final FileEntry itemToSelect = temp;
        if (isVisible()) {  // May have been canceled
            SwingUtilities.invokeLater(new Runnable() {

                public void run() {
                    if (!list.isEnabled()) {
                        list.setEnabled(true);
                    }
                    list.setModel(model);
                    if (model.size() > 0) {
                        int index = 0;
                        if (itemToSelect != null) {
                            FileItem fi;
                            String path = itemToSelect.getFullEscapedPath();
                            for (int i = 0; i < model.size(); i++) {
                                fi = (FileItem) model.get(i);
                                if (!fi.goUp && fi.entry.getFullEscapedPath().equals(path)) {
                                    index = i;
                                }
                            }
                        }
                        list.setSelectedIndex(index);
                        Rectangle r = list.getUI().getCellBounds(list, index, index);
                        if (r != null) {
                            r.height += r.height;
                            list.scrollRectToVisible(r);
                        }
                    } else {
                        valueChanged(null);
                    }
                }
            });
        }
    }

    private DefaultListModel populateList(FileItem root) {
        FileEntry entries[] = device.getFileListingService().getChildren(root.entry, true, null);
        DefaultListModel model = new DefaultListModel();
        if (entries != null) {
            List<FileItem> l = new ArrayList<FileItem>(entries.length);
            for (FileEntry entry : entries) {
                l.add(new FileItem(entry, false));
            }
            Collections.sort(l);
            if (root.entry.getParent() != null) {  // Add the "Go Up" item except for the root
                model.add(0, root);
            }
            for (FileItem fi : l) {
                model.add(model.size(), fi);
            }
        } else {
            model.add(0, root);
        }
        return model;
    }

    public void valueChanged(ListSelectionEvent e) {
        Object o = list.getSelectedValue();
        boolean enabled = list.isEnabled() && o instanceof FileItem && !((FileItem) o).goUp;
        btnOk.setEnabled(enabled);
    }

    public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() == 2) {
            FileItem sel = (FileItem) list.getSelectedValue();
            if (sel != null && sel.entry.isDirectory()) {
                if (sel.goUp) {  // Go one level up
                    FileEntry parent = sel.entry.getParent();
                    currentRoot = new FileItem(parent, true);
                    fileToSelect = sel.entry.getFullEscapedPath();
                } else {
                    currentRoot = new FileItem(sel.entry, true);
                }
                new Thread(this).start();
            }
        }
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    /**
     * Get the selected file.
     * @return the selected file or null if the window was canceled.
     */
    public String getSelectedFile() {
        return selectedFile;
    }

    /**
     * Set the selected file.
     * @param the file to select.
     */
    public void setSelectedFile(String fileToSelect) {
        this.fileToSelect = fileToSelect;
    }

    private class FileItem implements Comparable<FileItem> {

        FileItem(FileEntry entry, boolean goUp) {
            this.entry = entry;
            this.goUp = goUp;
        }
        final FileEntry entry;
        final boolean goUp;

        /**
         * Compare two FileItem instances. This ensures that the list will be
         * sorted in a way that the ".." directory is first followed by alphabetically
         * sorted directories followed by alphabetically sorted files.
         * @param o another file item.
         * @return -1 if the file is smaller (prior to) the argument one, 1 if
         * it is greater or 0 if it is equal.
         */
        public int compareTo(FileItem o) {
            if (goUp) { // The "Go Up" item is always listed as the first one
                return -1;
            }
            boolean d1 = entry.isDirectory();
            boolean d2 = o.entry.isDirectory();
            if (d1 && !d2) {
                return -1;
            } else if (!d1 && d2) {
                return 1;
            }
            return entry.getName().compareTo(o.entry.getName());
        }

        @Override
        public String toString() {
            return goUp ? ".." : entry.getName();
        }
    }

    private class FileRenderer extends DefaultListCellRenderer {

        @Override
        public Component getListCellRendererComponent(
                JList list,
                Object value,
                int index,
                boolean isSelected,
                boolean cellHasFocus) {
            JLabel c = (JLabel) super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

            if (value instanceof FileItem) {
                FileItem fi = (FileItem) value;
                Font f = c.getFont();
                if (fi.entry.isDirectory()) {
                    f = new Font(f.getName(), Font.BOLD, f.getSize());
                    c.setIcon(dirIcon);
                } else {
                    f = new Font(f.getName(), Font.PLAIN, f.getSize());
                    c.setIcon(fileIcon);
                }
                c.setFont(f);
            }

            return c;
        }
    }
}
