/*
 * T-Plan Robot Enterprise, automated testing tool.
 * Copyright (C) 2009-2013 T-Plan Limited (http://www.t-plan.com),
 * Cloisters House, 8 Battersea Park Road, London, SW8 4BG, United Kingdom.
 * All rights reserved.
 */
package com.tplan.adb;

// Import the ADB loader to ensure that the libs get loaded
import com.tplan.adb.AdbLoader;

import com.android.ddmlib.IDevice;
import com.android.ddmlib.ShellCommandUnresponsiveException;
import com.tplan.robot.ApplicationSupport;
import com.tplan.robot.gui.components.FileChooserManager;
import com.tplan.robot.preferences.Preference;
import com.tplan.robot.preferences.UserConfiguration;
import com.tplan.robot.remoteclient.RemoteDesktopClient;
import com.tplan.robot.remoteclient.adb.AdbClient;
import com.tplan.robot.scripting.DefaultJavaTestScript;
import com.tplan.robot.scripting.ParametrizedTestScript;
import com.tplan.robot.scripting.StopRequestException;
import com.tplan.robot.util.Utils;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.datatransfer.StringSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeoutException;
import java.util.logging.Level;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.border.Border;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * Install an application from the local .apk file on the Android device.
 */
public class Install extends DefaultJavaTestScript implements Constants,
        ParametrizedTestScript, Runnable, FileChooserManager, ActionListener {

    // Parameter "file"
    public static final String PARAM_FILE = "file";
    // Parameter "reinstall"
    public static final String PARAM_REINSTALL = "reinstall";
    // Variable to store the installed package to
    public static final String VAR_PACKAGE = "_ADB_INSTALLED_PACKAGE";
    // Variable to store the installed package to (since v2)
    public static final String VAR_ACTIVITY = "_ADB_INSTALLED_ACTIVITY";

    //
    // Member variables populated per execution
    protected IDevice device;
    protected String file;
    protected Exception exception;
    protected boolean reinstall;
    // Cached last .apk file to reopen the file chooser in the same location
    static File lastAppFile = null;

    // Cached last package and application (since v2)
    static String lastPackage = null;
    static String lastActivity = null;

    @Override
    public void test() {
        int exitCode = 1;
        try {
            // Load the command parameters
            final String file = getVariableAsString(PARAM_FILE);

            Boolean b = getContext().getVariableAsBoolean(PARAM_REINSTALL);
            this.reinstall = b != null && b;

            // There must be a connected "Android over ADB" client
            RemoteDesktopClient client = getContext().getClient();
            if (client != null && client.isConnected() && client instanceof AdbClient) {
                AdbClient cl = (AdbClient) client;
                IDevice dev = (IDevice) cl.getIDevice();
                if (dev != null) {
                    try {
                        // Populate the member variables
                        this.device = dev;
                        this.file = file;
                        this.exception = null;

                        // Start the operation in a thread so that we can
                        // continue in case that the user stops the script.
                        // As the IDevice API doesn't allow to stop the operation
                        // we will just let the thread run on its own.
                        Thread t = new Thread(this);
                        t.start();

                        // Wait until the thread finishes. If the stop is requested
                        // cancel it.
                        while (t.isAlive()) {
                            if (getContext().getMasterInterpret().isStop()) {
                                t.interrupt();
                                throw new StopRequestException();
                            }
                            Thread.sleep(5);
                        }

                        // If the command executed in the thread threw an exception
                        // then rethrow it
                        if (exception != null) {
                            throw exception;
                        }

                        exitCode = 0;
                    } catch (Exception ex) {
                        if (ex instanceof ShellCommandUnresponsiveException || ex instanceof TimeoutException) {
                            log("Failed to install application " + file + " to the Android device: Operation timed out.", Level.SEVERE);
                        } else if (ex instanceof StopRequestException) {
                            log("The script was stopped while installing application " + file + " to the Android device. The installation result is not known.", Level.WARNING);
                        } else {
                            log("Failed to install application " + file + " to the Android device: Operation failed with " + ex.getClass().getName() + ": " + ex.getMessage(), Level.SEVERE);
                        }
                    }
                }
            } else {
                if (client == null || !client.isConnected()) {
                    log("Failed to install application " + file + " because no Android device is connected!", Level.SEVERE);
                } else {
                    log("Failed to install application " + file + " because the current connection ("
                            + client.getConnectString() + ") is not an Android Over ADB one!", Level.SEVERE);
                }
            }
        } catch (StopRequestException ex) {
            throw ex;
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            getContext().setVariable("_EXIT_CODE", exitCode);
        }
    }

    public List<String> getParameters() {
        return Arrays.asList(new String[]{PARAM_FILE, PARAM_REINSTALL});
    }

    public List getParameterValues(String parameterName) {
        if (PARAM_REINSTALL.equals(parameterName)) {
            return Arrays.asList(new String[]{"true", "false"});
        }
        return null;
    }

    public void setParameters(Map<String, String> paramsAndValues) {
    }

    public String getLongDescription() {
        return getShortDescription();
    }

    public String getShortDescription() {
        return "Install an Android application (.apk) on the device.";
    }

    public void run() {
        try {
            // v2 updates - also store the main application activity
            getContext().getVariables().remove(VAR_PACKAGE);
            getContext().remove(VAR_PACKAGE);
            lastPackage = null;
            getContext().getVariables().remove(VAR_ACTIVITY);
            getContext().remove(VAR_ACTIVITY);
            lastActivity = null;
            device.installPackage(file, reinstall);
            String s[] = readPackage(file);
            String pkg = s[0];
            if (pkg != null) {
                getContext().setVariable(VAR_PACKAGE, pkg);
                getContext().put(VAR_PACKAGE, pkg);
                lastPackage = pkg;
            }
            String act = s[1];
            if (act != null) {
                getContext().setVariable(VAR_ACTIVITY, act);
                getContext().put(VAR_ACTIVITY, act);
                lastActivity = act;
            }
        } catch (Exception ex) {
            this.exception = ex;
        }
    }

    public List<Preference> getVisualParameters() {
        List<Preference> l = new ArrayList<Preference>();
        String title = "ADB Install";
        String desc = "<html><body><a href=\""
                + ADB_INSTALL_HELP_URL + "\">Install</a> a locally stored Android application (.apk) onto the Android device connected<br/>"
                + "over the "
                + "<a href=\"" + ANDROID_OVER_ADB_URL + "\">Android Over ADB</a> connection. The application package name will be stored<br/>"
                + "to the <code>" + VAR_PACKAGE + "</code> variable which can be later used for uninstallation."
                + "<br/>For details see the <a href=\"" + PLUGIN_DOC_URL + "#" + getClass().getSimpleName().toLowerCase()
                + "\">" + getClass().getSimpleName() + " specification</a>."
                + "</body></html>";
        Preference p = new Preference(PARAM_FILE, Preference.TYPE_FILE, "Application file (.apk)", desc);
        p.setFileChooserManager(this);
        p.setPreferredContainerName(title);
        p.setAcceptEmptyValue(false);
        p.setMandatory(true);
        p.setHint(title,
                "<html><body>"
                + "<p>Provide the full path of the application file (.apk) in the local file system. To specify the file relatively "
                + "to one of the standard paths use one of the default path variables such as:</p>"
                + "<ul><li><code>_INSTALL_DIR</code> - the Robot installation directory,</li>"
                + "<li><code>_PROJECT_DIR</code> - home folder of the project the script belongs to,</li>"
                + "<li><code>_SCRIPT_DIR</code> - the script folder (TPR scripts only).</li>"
                + "</ul><p>For example, to install the <code>myapp.apk</code> file stored in the project home folder "
                + "enter:</p><br/>"
                + "<ul><code>\"{_PROJECT_DIR}" + File.separator + "myapp.apk\"</code></ul>"
                + "</body></html>"
        );
        l.add(p);

        p = new Preference(PARAM_REINSTALL, Preference.TYPE_BOOLEAN, "Reinstall the app if already installed", null);
        p.setPreferredContainerName(title);
        p.setDefaultValue(false);
        p.setHint(title,
                "<html><body>"
                + "<p>Select this option if you want to reinstall the application "
                + "even if the same version is already installed on the device. The default value is "
                + "<code>false</code> (off; do not reinstall)."
                + "</body></html>"
        );
        l.add(p);
        return l;
    }

    public void setFileChooser(JFileChooser chooser) {
        chooser.setFileFilter(new FileNameExtensionFilter("Android applications (*.apk)", "apk"));
        chooser.addActionListener(this);
        if (lastAppFile != null) {
            chooser.setCurrentDirectory(lastAppFile.getParentFile());
            chooser.setSelectedFile(lastAppFile);
        }
        chooser.setAccessory(new Accessory(chooser));
    }

    public void actionPerformed(ActionEvent e) {
        if (e.getSource() instanceof JFileChooser) {
            lastAppFile = ((JFileChooser) e.getSource()).getSelectedFile();
        }
    }

    /**
     * Read the package name from the .apk file.
     *
     * @param file the .apk file.
     * @return the package name.
     * @throws IOException on an I/O error.
     */
    static String[] readPackage(String file) throws IOException {
        String s[] = readPackageThroughAAPT(file);
        if (s == null) {
            ZipFile zf = null;
            try {
                // The .apk is a zip file
                zf = new ZipFile(file);
                Enumeration e = zf.entries();
                ZipEntry ze;
                while (e.hasMoreElements()) {
                    ze = (ZipEntry) e.nextElement();
                    // The package is contained in the AndroidManifest.xml
                    if ("AndroidManifest.xml".equals(ze.getName())) {
                        long size = ze.getSize();
                        int initSize = (int) Math.min(size + 1, 4096);
                        ByteArrayOutputStream out = new ByteArrayOutputStream(initSize);
                        BufferedInputStream bin = new BufferedInputStream(zf.getInputStream(ze));
                        byte[] buffer = new byte[initSize];
                        int len;
                        while ((len = bin.read(buffer)) != -1) {
                            out.write(buffer, 0, len);
                        }
                        out.close();
                        return retrievePackage(out.toString());
                    }
                }
            } finally {
                if (zf != null) {
                    try {
                        zf.close();
                    } catch (IOException e1) {
                    }
                }
            }
            return null;
        }
        return s;
    }

    /**
     * Read the application package and main activity using aapt.
     * @param file the application file (.apk).
     * @return the data or null if aapt was not found or failed to recognize the 
     * file.
     * @since 2.0
     */
    private static String[] readPackageThroughAAPT(String file) {
        File aapt = getAapt();
        if (aapt != null) {
            List<String> cmd = new ArrayList<String>();
            cmd.add(aapt.getAbsolutePath());
            cmd.add("dump");
            cmd.add("badging");
            cmd.add(file);
            
            StringBuilder out = new StringBuilder();
            try {
                OSProcessThread thread = new OSProcessThread(cmd, out, null);
                long maxTime = System.currentTimeMillis() + 3000;
                thread.start();
                while (thread.isAlive()) {
                    Thread.sleep(2);
                    if (System.currentTimeMillis() >= maxTime) {
                        thread.interrupt();
                        return null;
                    }
                }
                
                // Process the output line by line
//                System.out.println(out);
                String ret[] = new String[2];
                BufferedReader r = new BufferedReader(new StringReader(out.toString()));
                String line;
                while ((line = r.readLine()) != null) {
                    line = line.trim();
                    if (line.startsWith("package:")) {
                        int index = line.indexOf("name='");
                        if (index > 0) {
                            line = line.substring(index+"name='".length());
                            index = line.indexOf("'");
                            if (index >= 0) {
                                line = line.substring(0, index);
                                ret[0] = line;
                            }
                        }
                    } else if (line.startsWith("launchable-activity:")) {
                        int index = line.indexOf("name='");
                        if (index > 0) {
                            line = line.substring(index+"name='".length());
                            index = line.indexOf("'");
                            if (index >= 0) {
                                line = line.substring(0, index);
                                ret[1] = line;
                            }
                        }
                    }
                    if (ret[0] != null && ret[1] != null) {
                        break;
                    }
                }
                return ret;
                
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        return null;
    }

    /**
     * Find the aapt binary in the Android SDK folder.
     * @return the aapt binary or null if not found.
     * @since 2.0
     */
    private static File getAapt() {
        String sdkPath = UserConfiguration.getInstance().getString("adb.path");
        if (sdkPath != null) {
            File dir = new File(sdkPath);
            if (dir.isDirectory()) {
                String name = System.getProperty("os.name").contains("Windows") ? "aapt.exe" : "aapt";
                File f = new File(dir, "build-tools");

                // The build-tools/ folder contains one or more subfolders with 
                // version specific set of build tools. As the naming convention
                // seems to be unstable we simply pick up the newest one which 
                // is likely to be the most recent Android version.
                File dirs[] = f.listFiles();
                if (dirs != null) {
                    File newest = null;
                    for (File ff : dirs) {
                        if (ff.isDirectory()) {
                            if (newest == null || newest.lastModified() < ff.lastModified()) {
                                newest = ff;
                            }
                        } else if (ff.getName().equals(name) && ff.canExecute()) {
                            // aapt placed directly under build-tools (unlikely)
                            return ff;
                        }
                    }
                    if (newest != null) {
                        f = new File(newest, name);
                        if (f.canExecute()) {
                            return f;
                        }
                    }
                }

                // The very last resort - search the folder for aapt
                List<File> l = Utils.findFile(dir, name, null, false, false, null, null);
                if (l.size() > 0) {
                    for (File ff : l) {
                        if (ff.canExecute()) {
                            return ff;
                        }
                    }
                }
            }
        }
        return null;
    }

    private static String filterText(String xml) {
        // The content is a binary file which contains various readable string
        // constants with individual characters separated by a space.
        // First of all filter out all characters which are not
        // a letter, digit, undescore '_' and period '.'.
        // Also add a space for every double space found between words.
        StringBuilder buf = new StringBuilder();
        int spaceCount = 0;
        for (char c : xml.toCharArray()) {
            if (Character.isLetter(c) || c == '.' || c == '_' || Character.isDigit(c)) {
                buf.append(c);
                spaceCount = 0;
            } else {  // Any other character is viewed as a separator
                if (spaceCount == 1 && buf.length() > 0) {  // Two characters -> add a space
                    buf.append(' ');
                }
                spaceCount++;
            }
        }
        return buf.toString();
    }

    /**
     * Retrieve the package from the content of the AndroidManifest.xml file.
     *
     * @param xml the file content.
     * @return the application package name.
     */
    private static String[] retrievePackage(String xml) {
        // The package name is typically after the "packagemanifest" prefix.
        String content = filterText(xml);
        String prefix = "package manifest ";
        int index = content.indexOf(prefix);
        String ret[] = new String[2];
        if (index >= 0) {
            // Read the string after the prefix until the next space.
            content = content.substring(index + prefix.length());
            StringBuilder buf = new StringBuilder();
            for (char c : content.toCharArray()) {
                if (Character.isLetter(c) || c == '.' || c == '_') {
                    buf.append(c);
                } else {
                    break;
                }
            }
            ret[0] = buf.toString();
        }

        prefix = " activity ";
        index = content.indexOf(prefix);
        if (index >= 0) {
            // Read the string after the prefix until the next space.
            content = content.substring(index + prefix.length());
            index = content.lastIndexOf("android.intent.action.MAIN");
            if (index >= 0) {
                content = content.substring(0, index);
                // First search for an activity starting with the package name
                index = ret[0] != null ? content.lastIndexOf(ret[0]) : -1;
                if (index >= 0) {
                    content = content.substring(index);
                } else {
                    // Return the first activity
                }
                StringBuilder buf = new StringBuilder();
                for (char c : content.toCharArray()) {
                    if (Character.isLetter(c) || c == '.' || c == '_') {
                        buf.append(c);
                    } else {
                        break;
                    }
                }
                ret[1] = buf.toString();
            }
        }
        return ret;
    }

    // Accessory component to display the package and main activity loaded 
    // from an application file displayed in a JFileChooser
    // Since v2
    private class Accessory extends JPanel implements PropertyChangeListener, Runnable, ActionListener {

        JLabel pkg = new JLabel();
        JLabel act = new JLabel();
        File file;
        Border border;
        JButton btnCopy = new JButton();

        public Accessory(JFileChooser owner) {
            setLayout(new GridBagLayout());
            add(new JLabel("Package: "), new GridBagConstraints(0, 0, 1, 1, 1.0, 0.0, GridBagConstraints.WEST,
                    GridBagConstraints.HORIZONTAL, new Insets(5, 5, 5, 5), 0, 0));
            add(pkg, new GridBagConstraints(0, 1, 1, 1, 1.0, 0.0, GridBagConstraints.WEST,
                    GridBagConstraints.HORIZONTAL, new Insets(5, 15, 5, 5), 0, 0));
            add(new JLabel("Main Activity: "), new GridBagConstraints(0, 2, 1, 1, 1.0, 0.0, GridBagConstraints.WEST,
                    GridBagConstraints.HORIZONTAL, new Insets(15, 5, 5, 5), 0, 0));
            add(act, new GridBagConstraints(0, 3, 1, 1, 1.0, 0.0, GridBagConstraints.WEST,
                    GridBagConstraints.HORIZONTAL, new Insets(5, 15, 5, 5), 0, 0));
            add(btnCopy, new GridBagConstraints(0, 4, 1, 1, 0.0, 0.0, GridBagConstraints.CENTER,
                    GridBagConstraints.NONE, new Insets(5, 15, 5, 5), 0, 0));
            border = BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Application Details");
            setBorder(border);
            btnCopy.setText("Copy");
            btnCopy.setToolTipText("Copy the package name and activity to the clipboard");
            btnCopy.setIcon(ApplicationSupport.getImageIcon("copy16.png"));
            btnCopy.addActionListener(this);
            owner.addPropertyChangeListener(this);
        }

        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            String prop = evt.getPropertyName();
            if (JFileChooser.DIRECTORY_CHANGED_PROPERTY.equals(prop)) {
                file = null;
            } else if (JFileChooser.SELECTED_FILE_CHANGED_PROPERTY.equals(prop)) {
                file = (File) evt.getNewValue();
            }
            if (file != null) {
                setVisible(true);
                pkg.setText("Reading " + file.getName() + "...");
                btnCopy.setEnabled(false);
                new Thread(Accessory.this).start();
            } else {
                setVisible(false);
                pkg.setText(null);
                act.setText(null);
            }
        }

        @Override
        public void run() {
            File f = this.file;
            if (f != null) {
                String p = null;
                String a = null;
                try {
                    String s[] = readPackage(f.getAbsolutePath());
                    File f2 = this.file;
                    if (f.equals(f2)) {
                        p = s[0];
                        a = s[1];
                    }
                } catch (Exception ex) {
                }
                final String pkg2 = p;
                final String act2 = a;
                SwingUtilities.invokeLater(new Runnable() {

                    @Override
                    public void run() {
                        pkg.setText(pkg2);
                        act.setText(act2);
                        btnCopy.setEnabled(pkg2 != null);
                    }
                });
            }
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            String s = pkg.getText() + "/" + act.getText();
            StringSelection stringSelection = new StringSelection(s);
            Toolkit.getDefaultToolkit().getSystemClipboard().setContents(stringSelection, null);
        }
    }
}
